import re

import numpy as np

from ..gl_importer import gl

__all__ = ['UniformDescription']


class UniformDescription:

    def __init__(self, name, array_length, is_matrix, method_name, dtype):
        self.name = name
        self.array_length = array_length
        self.is_matrix = is_matrix
        self.method_name = method_name
        self.dtype = dtype

    # New methods -------------------------------------------------------------
    def create_method(self, program_index):
        index = gl.glGetUniformLocation(program_index, self.name)
        method = getattr(gl, 'gl' + self.method_name)

        def internal_create_method(index, method, array_length):
            # pylint: disable=cell-var-from-loop
            if self.is_matrix:
                def h(ref):
                    # assert ref.dtype == h.description.dtype
                    method(index, array_length, gl.GL_TRUE, ref)
            elif self.array_length is not None:
                def h(ref):
                    # assert ref.dtype == h.description.dtype
                    method(index, array_length, ref)
            else:
                def h(value):
                    # assert value.dtype == h.description.dtype
                    method(index, value)
            h.description = self
            return h
        return internal_create_method(index, method, self.array_length)

    @classmethod
    def parse_shader_text(cls, shader_text):
        """
        Parses the shader text and returns an iterable of UniformDescription
        objects.
        """
        for type_, name, array in re.findall(
                r'^uniform (\w*|unsigned int) (\w*)(?:\[(\d*)\]|);',
                shader_text,
                re.MULTILINE):

            array_length = int(array) if array else None

            u = (cls.parse_vector(name, type_, array_length)
                 or cls.parse_matrix(name, type_, array_length)
                 or cls.parse_scalar(name, type_, array_length))
            if u is None:
                raise ValueError(f"Type {type_} not understood")
            yield u

    @classmethod
    def parse_vector(cls, name, type_, array_length):
        """
        Returns a UniformDescription given parsed values:
        * name, a string,
        * type_, a string such as bvec4, and
        * array_length, an integer or None.
        """
        # Check if type_ is a vector.
        m = re.match('(i|u|d|b|)vec(.)', type_)
        if not m:
            return None
        type_code, shape = m.groups()
        base_type = cls._translate_type_code(type_code)
        return UniformDescription(name,
                                  array_length or 1,
                                  False,
                                  f'Uniform{shape}{base_type}v',
                                  cls._dtype_of_base_type(base_type))

    @classmethod
    def parse_matrix(cls, name, type_, array_length):
        """
        Returns a UniformDescription given parsed values:
        * name, a string,
        * type_, a string such as bvec4, and
        * array_length, an integer or None.
        """
        m = re.match('(d|)mat(.*)$', type_)
        if not m:
            return None
        type_code, shape = m.groups()
        base_type = cls._translate_type_code(type_code)
        return UniformDescription(name,
                                  array_length or 1,
                                  True,
                                  f'UniformMatrix{shape}{base_type}v',
                                  cls._dtype_of_base_type(base_type))

    @classmethod
    def parse_scalar(cls, name, type_, array_length):
        """
        Returns a UniformDescription given parsed values:
        * name, a string,
        * type_, a string such as bvec4, and
        * array_length, an integer or None.
        """
        if any(type_.startswith(x + 'sampler')
               for x in ['', 'i', 'u']):
            type_code = 'i'
        else:
            try:
                type_code = {'bool': 'b',
                             'int': 'i',
                             'uint': 'ui',
                             'unsigned int': 'ui',
                             'float': 'f',
                             'double': 'd'}[type_]
            except KeyError:
                return None
        base_type = cls._translate_type_code(type_code)
        return UniformDescription(
            name,
            array_length,
            False,
            f'Uniform1{base_type}{"v" if array_length else ""}',
            cls._dtype_of_base_type(base_type))

    # Private methods ---------------------------------------------------------
    @classmethod
    def _translate_type_code(cls, type_code):
        return {'': 'f',
                'b': 'ui',
                'u': 'ui'}.get(type_code, type_code)

    @classmethod
    def _dtype_of_base_type(cls, base_type):
        return {'i': np.int32,
                'ui': np.uint32,
                'f': np.float32,
                'd': np.float64}[base_type]



import numpy as np
from rectangle import Rect

from ..transformations import clip_matrix

__all__ = ['OrthoProjection']


class OrthoProjection:
    """
    A projection coverts between two spaces:
    * widget (determined externally by the widget_rect)
    * gl (the gl coördinates, which range from -1 to 1 on each axis)
    """

    def __init__(self, widget_rect=Rect(sizes=[1, 1])):
        """
        widget_rect
        * shape: (2,)
        * in screen coordinates
        """
        self.widget_rect = widget_rect

    # Matrix properties:
    @property
    def widget_to_gl(self):
        """
        This is the projection matrix.
        """
        wr = self.widget_rect
        return (np.diag([1, -1, 1, 1])
                @ clip_matrix(wr.mins[0], wr.maxes[0],
                              wr.mins[1], wr.maxes[1],
                              -1.0, 1.0))

    @property
    def gl_to_widget(self):
        """
        This is the inverse projection matrix.
        """
        return np.linalg.inv(self.widget_to_gl)

    def __repr__(self):
        return f"{type(self).__name__}(widget_rect={self.widget_rect})"



import numpy as np

from ..transformations import translation_matrix

__all__ = ['OrthoView']


class OrthoView:
    """
    An OrthoView converts between four spaces:
    * scene
    * screen (the scene space zoomed so that one unit is one pixel)
    * widget (the screen space translated so that the origin is in the
      top-left)
    * scaled_widget (device_pixel_ratio times smaller than widget), which
      accounts for a scaling trick on retina screens in OS X.
    """

    def __init__(self, zoom=np.ones(2), scroll=np.zeros(2, dtype='i')):
        """
        * zoom
          * shape: (2,)
          * Conversion from scene space to screen space.
        * scroll
          * shape: (2,)
          * Conversion from screen space to widget space.
          * Bottom-right is positive.
        """
        super().__init__()
        # These variables do the transformation between spaces.
        self.zoom_ = np.ones(2) * zoom
        self.scroll_ = np.array(scroll, dtype='i')
        self.device_pixel_ratio = 1.0
        assert self.scroll.shape == (2,)

    # Settable vector properties ----------------------------------------------
    @property
    def zoom(self):
        return self.zoom_

    @zoom.setter
    def zoom(self, new_zoom):
        self.zoom_ = new_zoom

    @property
    def scroll(self):
        return self.scroll_

    @scroll.setter
    def scroll(self, new_scroll):
        self.scroll_ = new_scroll

    # Vector properties -------------------------------------------------------
    @property
    def flipped_zoom(self):
        return np.array([self.zoom_[0], -self.zoom_[1]])

    # Matrix properties -------------------------------------------------------
    @property
    def scene_to_screen(self):
        return np.diag([self.zoom[0], -self.zoom[1], 1.0, 1.0])

    @property
    def widget_to_screen(self):
        return translation_matrix([self.scroll_[0],
                                   self.scroll_[1],
                                   0.0])

    @property
    def widget_to_scene(self):
        """
        The inverse view matrix.
        """
        return self.screen_to_scene.dot(self.widget_to_screen)

    @property
    def screen_to_scene(self):
        return np.diag([1 / self.zoom[0], -1 / self.zoom[1], 1.0, 1.0])

    @property
    def screen_to_widget(self):
        return np.linalg.inv(self.widget_to_screen)

    @property
    def scene_to_widget(self):
        """
        This is the view matrix.
        """
        return self.screen_to_widget.dot(self.scene_to_screen)

    @property
    def scaled_widget_to_scene(self):
        return self.widget_to_scene.dot(self.scaled_widget_to_widget)

    @property
    def scaled_widget_to_screen(self):
        return self.widget_to_screen.dot(self.scaled_widget_to_widget)

    @property
    def scaled_widget_to_widget(self):
        return np.diag([self.device_pixel_ratio, self.device_pixel_ratio,
                        1.0, 1.0])



from .gl_importer import gl

__all__ = ['glGetActiveAttrib']


def glGetActiveAttrib(program, index):
    """Wrap PyOpenGL glGetActiveAttrib as for glGetActiveUniform
    """
    buffer_size, = gl.glGetProgramiv(program,
                                     gl.GL_ACTIVE_ATTRIBUTE_MAX_LENGTH)
    length = gl.GLsizei()
    size = gl.GLint()
    type_ = gl.GLenum()
    name = bytes(buffer_size)

    gl.glGetActiveAttrib(program, index, buffer_size,
                         length, size, type_, name)
    return name.decode().rstrip('\x00'), size.value, type_.value



from io import StringIO

from mako.runtime import Context
from mako.template import Template

from ..gl_importer import gl as gl

__all__ = ['Shader']


class Shader:
    EXTENSION = {
        gl.GL_VERTEX_SHADER: 'vert',
        gl.GL_GEOMETRY_SHADER: 'geom',
        gl.GL_FRAGMENT_SHADER: 'frag',
        gl.GL_COMPUTE_SHADER: 'comp'}

    def __init__(self, filename, type_, context_kwargs=None):
        self.type_ = type_

        template = Template(filename=filename)
        buffer = StringIO()
        context = Context(buffer, **(context_kwargs
                                     if context_kwargs is not None
                                     else {}))
        try:
            template.render_context(context)
        except:
            from mako import exceptions
            print(exceptions.text_error_template().render())
        shader_text = buffer.getvalue()

        # pylint: disable=assignment-from-no-return
        self.shader_index = gl.glCreateShader(type_)
        assert self.shader_index

        gl.glShaderSource(self.shader_index, shader_text)
        gl.glCompileShader(self.shader_index)
        result = gl.glGetShaderiv(self.shader_index, gl.GL_COMPILE_STATUS)
        if result != 1:
            log = gl.glGetShaderInfoLog(self.shader_index).decode('latin')
            print(shader_text)
            raise Exception("""
                            Couldn't compile shader.
                            Shader compilation log:
                            """ + log)
        self.shader_text = shader_text

    def delete_shader(self):
        gl.glDeleteShader(self.shader_index)
        self.shader_index = 0



import freetype as ft
import numpy as np

from ..gl_importer import gl as gl
from ..shader_program import Attribute, BufferDescription

__all__ = ['DisplayList']


class DisplayList:
    RECORD_TYPE = np.dtype([('vertex', '<f4', 2),
                            ('code', '<i4')])

    def __init__(self, font):
        self.font = font
        self.buffer, = gl.glGenBuffers(1)
        self.vertex_array, = font.shader_program.create_vertex_arrays(
            [BufferDescription(
                self.buffer,
                self.RECORD_TYPE,
                [Attribute('vertex', ['vertex'], is_vector=True),
                 Attribute('code', ['code'])])])
        self.array = None
        self.text = None
        self.characters = {}

    def delete(self):
        # Work around the fact that glGenBuffers returns a result that
        # glDeleteBuffers can't handle.
        gl.glDeleteBuffers(1, int(self.buffer))
        self.vertex_array.delete()

    def set_text(self, text):
        """
        Fill the buffer and vertex array.
        Possibly render additional glyphs into the texture.
        """
        if not isinstance(text, str):
            raise TypeError("text argument must be a string — not {}".format(
                type(text)))
        if (text == self.text
                and self.atlas_creation_id == self.font.atlas.creation_id):
            return
        self.text = text
        self.regenerate()

    def regenerate(self):
        face = self.font.face
        options = self.font.FT_OPTIONS

        self.atlas_creation_id = self.font.atlas.creation_id
        self.array = np.empty(len(self.text), self.RECORD_TYPE)
        vertex = np.zeros(2, dtype='f')
        last_char_index = 0

        for i, c in enumerate(self.text):
            this_vertex = self.array[i: i + 1]

            # Set code.
            this_vertex['code'] = self.font.get_char(c)

            # get_char might have caused creation_id to change.
            if self.atlas_creation_id != self.font.atlas.creation_id:
                self.regenerate()
                return

            # Set vertex.
            face.load_char(c, options)
            glyph = face.glyph

            this_char_index = face.get_char_index(c)
            assert this_char_index > 0

            kerning = face.get_kerning(last_char_index, this_char_index,
                                       ft.FT_KERNING_DEFAULT)

            advance = glyph.linearHoriAdvance / 65536
#           print("==========")
#           print(glyph.format)
#           print(c)
#           print("glyph", glyph.bitmap.rows)
#           for x in ['linearHoriAdvance', 'linearVertAdvance',
#                     'bitmap_left', 'bitmap_top']:
#               print(x, getattr(glyph, x))
#           print("glyph metrics")
#           for x in ['horiAdvance', 'horiBearingX', 'horiBearingY',
#                     'vertAdvance', 'vertBearingX', 'vertBearingY', 'width']:
#               print(x, getattr(glyph.metrics, x))
            bottom_left = np.array(
                (glyph.bitmap_left + kerning.x / 64,
                 -glyph.bitmap_top + glyph.bitmap.rows + kerning.y / 64),
                dtype='f')

            this_vertex['vertex'] = vertex + bottom_left
            vertex[0] += advance

            last_char_index = this_char_index

        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, self.buffer)
        gl.glBufferData(gl.GL_ARRAY_BUFFER,
                        self.array.nbytes,
                        self.array,
                        gl.GL_STATIC_DRAW)
        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, 0)

    def draw(self, widget_point):
        """
        Before drawing, be sure to set the uniforms:
        * projection
        * color
        * gamma
        * scale
        """
        if self.atlas_creation_id != self.font.atlas.creation_id:
            self.regenerate()
        with self.vertex_array.bind_context():
            self.font.shader_program.vertex_offset(widget_point)
            gl.glDrawArrays(gl.GL_POINTS, 0, len(self.text))



from contextlib import contextmanager

import numpy as np

from ..gl_importer import gl as gl
from ..tools import next_power_of_two

__all__ = ['CodeLookup']


class CodeLookup:

    def __init__(self, font, size, texture_unit):
        self.clear()
        self.resize(size)
        self.texture_unit = texture_unit
        self.font = font

        with self.font.shader_program.bind_context():
            # Create sampler.
            self.sampler_object, = gl.glGenSamplers(1)

            # Create texture.
            self.texture, = gl.glGenTextures(1)

            # Bind sampler to texture unit, and set parameters.
            gl.glActiveTexture(gl.GL_TEXTURE0 + self.texture_unit)
            gl.glBindSampler(self.texture_unit, self.sampler_object)
            gl.glSamplerParameteri(self.sampler_object,
                                   gl.GL_TEXTURE_WRAP_S, gl.GL_CLAMP_TO_EDGE)
            gl.glSamplerParameteri(self.sampler_object,
                                   gl.GL_TEXTURE_MAG_FILTER, gl.GL_NEAREST)
            gl.glSamplerParameteri(self.sampler_object,
                                   gl.GL_TEXTURE_MIN_FILTER, gl.GL_NEAREST)

            # Set uniform with texture unit.
            self.font.shader_program.code_to_texture(
                np.int32(self.texture_unit))

    def resize(self, size):
        self.data = np.empty((next_power_of_two(size), 4), dtype='f')
        self.needs_update = True

    def clear(self):
        self.used = 0
        self.needs_update = True

    @property
    def size(self):
        return self.data.shape[0]

    def add_char(self, uvs):
        if self.used == self.size:
            old_data = self.data
            old_size = self.size
            self.resize(2 * self.size)
            self.data[0: old_size] = old_data
        self.data[self.used] = uvs
        self.used += 1
        self.needs_update = True

    def update_texture(self):
        # Bind texture to texture unit, set paramters and upload texture.
        # ActiveTexture must precede TexParameter, BindTexture,
        # and TexImage.
        gl.glActiveTexture(gl.GL_TEXTURE0 + self.texture_unit)
        gl.glBindTexture(gl.GL_TEXTURE_1D, self.texture)
        gl.glTexImage1D(gl.GL_TEXTURE_1D,
                        0,
                        gl.GL_RGBA32F,
                        self.data.shape[0],
                        0,  # border
                        gl.GL_RGBA,
                        gl.GL_FLOAT,
                        self.data)
        self.needs_update = False

    @contextmanager
    def draw_context(self):
        gl.glActiveTexture(gl.GL_TEXTURE0 + self.texture_unit)
        gl.glBindTexture(gl.GL_TEXTURE_1D, self.texture)
        gl.glBindSampler(self.texture_unit, self.sampler_object)
        yield
        gl.glActiveTexture(gl.GL_TEXTURE0 + self.texture_unit)
        gl.glBindTexture(gl.GL_TEXTURE_2D, 0)
        gl.glBindSampler(self.texture_unit, 0)



from rectangle import Rect

from ..transformations import four_components

__all__ = ['Viewport']


class Viewport:
    """
    A Viewport contains
    * a projection, which maps between OpenGL and widget coördinates, and
    * a view, which maps between widget and scene coördinates.

    Theoretically viewports should also have a location within the widget.
    """

    def __init__(self, projection, view):
        self.projection = projection
        self.view = view

    # New methods -------------------------------------------------------------
    def scene_visible_rect(self):
        """
        The widget rectangle mapped to scene space.
        """
        return self.projection.widget_rect.transformed(
            self.view.widget_to_scene).rectified()

    # Magic methods -----------------------------------------------------------
    def __repr__(self):
        return (f"{type(self).__name__}("
                f"projection={self.projection}, "
                f"view={self.view})")



import numpy as np

__all__ = ['four_components',
           'translation_matrix',
           'clip_matrix']


def four_components(x):
    "Pads x to four components using [0, 0, 1].  I.e., [2] -> [2, 0, 0, 1]."
    return np.concatenate((x, [0, 0, 1][len(x) - 1:]), axis=0)


def translation_matrix(direction):
    "Returns a matrix that translates in the given direction"
    direction = np.asarray(direction)
    if direction.shape != (3,):
        raise ValueError("direction has invalid shape")
    retval = np.identity(4)
    retval[:3, 3] = direction[:3]
    return retval


def clip_matrix(left, right, bottom, top, near, far, perspective=False):
    """Returns a matrix to obtain normalized device coordinates from a
    frustum.

    The frustum bounds are axis-aligned along x (left, right),
    y (bottom, top) and z (near, far).

    Normalized device coordinates are in the range [-1, 1] if the coordinates
    are inside the frustum.

    If perspective is true, the frustum is a truncated pyramid with the
    perspective point at origin and direction along z axis, otherwise it is an
    orthographic canonical view volume (a box).

    Homogeneous coordinates transformed by the perspective clip matrix
    need to be dehomogenized (divided by the w coordinate).
    """
    if left >= right:
        raise ValueError("invalid frustum: left ≥ right")
    if bottom >= top:
        raise ValueError("invalid frustum: bottom ≥ top")
    if near >= far:
        raise ValueError("invalid frustum: near ≥ far")
    if perspective:
        if near <= 1e-6:
            raise ValueError("invalid frustum: near <= 0")
        t = 2.0 * near
        m = [[t / (left - right), 0.0, (right + left) / (right - left), 0.0],
             [0.0, t / (bottom - top), (top + bottom) / (top - bottom), 0.0],
             [0.0, 0.0, (far + near) / (near - far), t * far / (far - near)],
             [0.0, 0.0, -1.0, 0.0]]
    else:
        m = [[2.0 / (right - left), 0.0, 0.0, (right + left) / (left - right)],
             [0.0, 2.0 / (top - bottom), 0.0, (top + bottom) / (bottom - top)],
             [0.0, 0.0, 2.0 / (far - near), (far + near) / (near - far)],
             [0.0, 0.0, 0.0, 1.0]]
    return np.array(m)



import numpy as np

__all__ = ['BufferDescription']


class BufferDescription:

    def __init__(self, buffer_index, buffer_dtype, attributes):
        """
        * buffer_index is a buffer object index (as returned by glGenBuffers).
        * buffer_dtype is the dtype of the numpy array that will be loaded
          into the buffer.
        * attributes is a reiterable of BoundAttribute objects.
        """
        if not isinstance(buffer_index, np.uint32):
            raise TypeError
        if not isinstance(buffer_dtype, np.dtype):
            raise TypeError

        self.buffer_index = buffer_index
        self.buffer_dtype = buffer_dtype
        self.attributes = attributes



# gl should be imported through this file so that the options are set before
# importing OpenGL.GL.
import OpenGL
OpenGL.SIZE_1_ARRAY_UNPACK = False
# ERROR_CHECKING = True
# ERROR_LOGGING = False
# ERROR_ON_COPY = False
# ARRAY_SIZE_CHECKING = True
# STORE_POINTERS = True
# WARN_ON_FORMAT_UNAVAILABLE = False
# FORWARD_COMPATIBLE_ONLY = False
# SIZE_1_ARRAY_UNPACK = True
# USE_ACCELERATE = True
# CONTEXT_CHECKING = False
#
# FULL_LOGGING = False
# ALLOW_NUMPY_SCALARS = False
# UNSIGNED_BYTE_IMAGES_AS_STRING = True
# MODULE_ANNOTATIONS = False
import OpenGL.GL as gl


__all__ = ['OpenGL', 'gl']



import ctypes
from itertools import count

import numpy as np

from ..gl_importer import OpenGL, gl

__all__ = []


class Indexer:

    def __init__(self, shape, strides):
        """
        Members
        * shape:
          The shapes of the indexed dimensions.
        * strides:
          The strides in bytes of the indexed dimensions.
        """
        self.shape = shape
        self.strides = strides

    def __repr__(self):
        return "Indexer(shape={}, strides={})".format(
            self.shape,
            self.strides)

    def get_offset(self, index):
        return int(np.dot(index, self.strides))


class BoundAttribute:

    """
    A BoundAttribute knows how to bind a single attribute in a shader.
    """

    def __init__(self, attribute, program, buffer_dtype):
        """
        The constructor accepts:
        * attribute: An Attribute object that describes the location of some
          data within a numpy array.
        * program: An instance of ShaderProgram.
        * buffer_dtype:  The type of the numpy array that represents the
          buffer.

        The constructor calculates the parameters of the OpenGL calls (e.g.,
        glVertexAttribPointer) that will need to be made when the
        BoundAttribute is bound.

        The constructor sets members relating to calling OpenGL functions:
        * attribute_location:  The location of the attribute in the program
          returned by glGetAttribLocation.
        * stride:
          The byte offset between consecutive “vertex attributes”.
        * offset:
          The offset in bytes into the buffer of the first “vertex attribute”.
        * integral:
          Whether the data in the buffer is integral or floating point.
        * gl_type:
          The GL type, e.g., GL_FLOAT or GL_INT
        * vector_size:
          The number of components per “vertex attribute”. Must be 1, 2, 3,
          or 4.
        * array_size:
          The number of OpenGL attributes represented by this BoundAttribute
          object.
        * array_stride:
          The stride in bytes between each component of the array.

        The constructor also sets members relating to indexing:
        * indexer:
          An instance of Indexer that translates from a sub-array index into
          the numpy array to a byte offset.  It also knows the maximum
          sub-array index for the purpose of assertions.  A sub-array is the
          leftover dimensions in the numpy array that are bound at call time.

        Let the shape of the buffer_dtype have n dimensions.  Then,
        m = n - is_vector - (array_size is not None) ≥ 0.  Iff m > 0, this is
        an indexed attribute.
        * indexed attributes need to be bound at draw time, e.g.,
          vertex_array.attributes[attrib_name].bind(index).
        * non-indexed attributes can be bound once to associate them with the
          vertex array (so binding the vertex array) automatically binds them.
        """
        self.attribute_name = attribute.name
        self.attribute_location = \
            program.attribute_name_to_location(attribute.name)
        self.stride = buffer_dtype.itemsize

        # Calculate offset and sub_dtype.
        self.offset = 0
        sub_dtype = buffer_dtype
        for lookup in attribute.lookup_sequence:
            if sub_dtype.fields is None:
                if not isinstance(lookup, int):
                    raise TypeError
                strides = np.zeros((), dtype=sub_dtype).strides
                if lookup > sub_dtype.shape[0]:
                    raise ValueError
                offset_delta = lookup * strides[0]
                sub_dtype = np.dtype((sub_dtype.base, sub_dtype.shape[1:]))
            else:
                if not isinstance(lookup, str):
                    raise TypeError
                sub_dtype, offset_delta = sub_dtype.fields[lookup]

            self.offset += offset_delta

        # Calculate shape and indexer.
        shape = sub_dtype.shape
        strides = np.zeros((), dtype=sub_dtype).strides

        def consume_last_dimension():
            nonlocal shape
            nonlocal strides
            # The last entry of shape and stride have to do with this
            # vector, which has now been consumed.
            shape = shape[:-1]
            strides = strides[:-1]

        if attribute.is_packed_array:
            if attribute.is_vector:
                raise ValueError
            if shape[-1] <= (attribute.array_size - 1) * 4:
                raise ValueError(
                    "The attribute's array_size is unnecessarily large given "
                    "the Numpy array shape's final component "
                    f"{shape[-1]}.")
            if attribute.array_size is None:
                raise ValueError(
                    "The array_size of a packed array is not specified")

            self.vector_size = 4
            self.array_size = attribute.array_size
            self.array_stride = strides[-1] * 4
            consume_last_dimension()
        else:
            if attribute.is_vector:
                if not (1 <= shape[-1] <= 4):
                    raise ValueError

                self.vector_size = shape[-1]
                consume_last_dimension()
            else:
                self.vector_size = 1

            if attribute.array_size is None:
                self.array_size = 1
                self.array_stride = 0
            else:
                if attribute.array_size != shape[-1]:
                    raise ValueError(
                        f'Attribute array is supposed to have size '
                        f'{attribute.array_size}, '
                        f'but the buffer has dimensions {shape}')
                self.array_size = attribute.array_size
                consume_last_dimension()

        if shape:  # Any leftover dimensions must be indexed at call time.
            self.indexer = Indexer(shape, strides)
        else:
            self.indexer = None

        # Calculate gl_type and integral.
        self.gl_type = OpenGL.arrays.numpymodule.ARRAY_TO_GL_TYPE_MAPPING[
            sub_dtype.base]
        self.integral = np.issubdtype(sub_dtype, np.integer)

    def enable_attributes(self):
        for i in range(self.array_size):
            gl.glEnableVertexAttribArray(self.attribute_location + i)

    def bind(self, index=None):
        """
        bind makes the OpenGL calls to specify the location and data format of
        the array of vertex attributes.
        * index is the index into the dimensions (that are not consumed by
          is_vertex or array_index) of the vertex buffer (a numpy array).
        """
        offset = self.offset
        if index is not None:
            try:
                offset += self.indexer.get_offset(index)
            except ValueError as e:
                raise IndexError(
                    f"Bad index {index} on attribute '{self.attribute_name}' "
                    f"that wants {len(self.indexer.strides)} components.") \
                    from e
        for attrib_index, this_offset in zip(
                range(self.array_size),
                count(offset, self.array_stride)):
            pointer = ctypes.c_void_p(this_offset)
            this_attrib_location = self.attribute_location + attrib_index
            if self.integral:
                gl.glVertexAttribIPointer(this_attrib_location,
                                          self.vector_size,
                                          self.gl_type,
                                          self.stride,
                                          pointer)
            else:
                gl.glVertexAttribPointer(this_attrib_location,
                                         self.vector_size,
                                         self.gl_type,
                                         gl.GL_FALSE,
                                         self.stride,
                                         pointer)

    def __str__(self):
        return ("BoundAttribute("
                + ", ".join("{}={}".format(x, getattr(self, x))
                            for x in ['attribute_location',
                                      'stride',
                                      'offset',
                                      'integral',
                                      'gl_type',
                                      'vector_size',
                                      'array_size',
                                      'array_stride',
                                      'indexer']) + ")")



from contextlib import contextmanager

from ..gl_importer import gl as gl
from .bound_attribute import BoundAttribute

__all__ = []


class VertexArray:

    """
    A VertexArray manages an OpenGL vertex array, which is a collection of
    bindings for the varying variables in a ShaderProgram.  Instances of
    BoundAttribute are created to do the individual binding.
    """

    def __init__(self, program, buffer_description):
        """
        * program is a ShaderProgram object.
        * buffer_description is a BufferDescription object.

        Creating this object will enable the specified attributes, which
        associates them with the bound buffer.
        """
        # The GL vertex array index.
        self.vertex_array = gl.glGenVertexArrays(1)
        # A map from attribute name to an instance of BoundAttribute.
        self.attributes = {}

        with self.bind_context():
            # Bind buffer for attribute name.
            gl.glBindBuffer(gl.GL_ARRAY_BUFFER,
                            buffer_description.buffer_index)

            for attribute in buffer_description.attributes:
                bound_attribute = BoundAttribute(
                    attribute,
                    program,
                    buffer_description.buffer_dtype)
                bound_attribute.enable_attributes()
                if not bound_attribute.indexer:
                    bound_attribute.bind()
                self.attributes[attribute.name] = bound_attribute

            # Finished describing buffers.
            gl.glBindBuffer(gl.GL_ARRAY_BUFFER, 0)

    def delete(self):
        assert self.vertex_array is not None
        gl.glDeleteVertexArrays(1, [self.vertex_array])
        self.vertex_array = None

    @contextmanager
    def bind_context(self):
        if self.vertex_array is None:
            yield
        else:
            gl.glBindVertexArray(self.vertex_array)
            try:
                yield
            finally:
                gl.glBindVertexArray(0)



from .font import *
from .gl_importer import *
from .shader_program import *
from .shader_programs import *
from .tools import *
from .transformations import *
from .viewport import *
from .wrappers import *



import string
from contextlib import contextmanager

import freetype as ft
import numpy as np
from pkg_resources import resource_filename

from ..shader_program import ShaderProgram
from .atlas import Atlas
from .code_lookup import CodeLookup

__all__ = ['Font']


class Font:
    ATLAS_TEXTURE_UNIT = 0
    CODE_TEXTURE_UNIT = 1
    DEFAULT_CHARACTERS = (string.ascii_letters
                          + string.digits
                          + string.punctuation
                          + ' ')
    FT_OPTIONS = (ft.FT_LOAD_RENDER
                  | ft.FT_LOAD_FORCE_AUTOHINT)
    # ft.FT_LOAD_TARGET_LCD

    def __init__(self, filename, size):
        """
        Create all textures for latin characters and store them on the card.
        Set uniforms for the textures.
        """
        self.shader_program = ShaderProgram(
            vertex=[resource_filename('glx', 'glsl_shaders/text.vert')],
            geometry=[resource_filename('glx', 'glsl_shaders/text.geom')],
            fragment=[resource_filename('glx', 'glsl_shaders/text.frag')])
        self.char_to_index = {}
        self.chars = []
        self.atlas = Atlas(self, 256, self.ATLAS_TEXTURE_UNIT)
        self.code_lookup = CodeLookup(self, len(self.DEFAULT_CHARACTERS),
                                      self.CODE_TEXTURE_UNIT)

        self.face = ft.Face(filename)
        self.face.set_char_size(size * 64)

        for c in self.DEFAULT_CHARACTERS:
            self.add_char(c)

        with self.shader_program.bind_context():
            self.shader_program.font_atlas(np.int32(self.atlas.texture_unit))
            self.shader_program.code_to_texture(np.int32(
                self.code_lookup.texture_unit))
            self.shader_program.gamma(np.float32(2.2))

    def add_char(self, c, register=True):
        self.face.load_char(c, self.FT_OPTIONS)

        glyph = self.face.glyph
        bitmap = glyph.bitmap

        uvs = self.atlas.add_char(bitmap)
        self.code_lookup.add_char(uvs)

        if register:
            self.char_to_index[c] = len(self.chars)
            self.chars.append(c)
            # self.display()

    def get_char(self, c):
        if c not in self.char_to_index:
            self.add_char(c)
        return self.char_to_index[c]

    def sort_chars(self):
        def sort_key(c):
            index = self.char_to_index[c]
            uvs = self.code_lookup.data[index]
            return uvs[1] - uvs[3]
        self.chars.sort(key=sort_key, reverse=False)
        self.char_to_index = {c: i
                              for i, c in enumerate(self.chars)}

    def repopulate(self):
        self.sort_chars()
        self.atlas.clear()
        self.code_lookup.clear()
        for c in self.chars:
            self.add_char(c, register=False)

    @contextmanager
    def draw_context(self, *args, **kwargs):
        if self.atlas.needs_update:
            self.atlas.update_texture()
        if self.code_lookup.needs_update:
            self.code_lookup.update_texture()
        with self.shader_program.bind_context(*args, **kwargs):
            with self.atlas.draw_context(), self.code_lookup.draw_context():
                yield

    def display(self):
        if self.atlas.lines:
            line = self.atlas.lines[-1]
            last_line_usage = line.x * line.height / self.atlas.size
        else:
            last_line_usage = 0
        for line in self.atlas.lines:
            print(line.height / self.atlas.size)
        print(self.atlas.size,
              (self.atlas.y + last_line_usage) / self.atlas.size)
        for c, d in zip(self.chars,
                        self.code_lookup.data[: self.code_lookup.used]):
            print(c, d)
        print('-')
        print()



import networkx as nx
import numpy as np
from numpy.testing import assert_allclose
from rectangle import Rect

from ..ortho_projection import OrthoProjection
from ..ortho_view import OrthoView
from ...transformations import four_components


def all_paths(g):
    """
    Returns a generator of all possible path lists such that each path list has
    the same start and end points.
    """
    for x in g:
        for y in g:
            paths = list(nx.all_simple_paths(g, x, y))
            if not paths:
                continue
            yield paths


def create_graph_from_space_converter(space_converter):
    methods = [(method, method.split('_to_'))
               for method in dir(type(space_converter))
               if '_to_' in method]

    g = nx.DiGraph()
    for method, path in methods:
        g.add_edge(*path, method=method)

    return g


def check_paths_consistent(g,
                           start_position,
                           space_converter,
                           paths,
                           iterate):
    # print("Checking", type(space_converter).__name__, paths)
    final_position = None
    for path in paths:
        position = start_position
        for edge_start, edge_end in zip(path, path[1:]):
            method = g[edge_start][edge_end]['method']
            position = iterate(position, getattr(space_converter, method))

        if final_position is None:
            final_position = position
            continue

        assert_allclose(final_position, position)


def process(space_converter, start_position_f, iterate,
            desired_total_path_lengths):
    g = create_graph_from_space_converter(space_converter)
    total_path_lengths = 0
    for paths in all_paths(g):
        total_path_lengths += len(paths)
        start_position = start_position_f(space_converter, paths[0][0])
        check_paths_consistent(g,
                               start_position,
                               space_converter,
                               paths,
                               iterate)
    assert total_path_lengths == desired_total_path_lengths


def iterate_rect(rect, t):
    return rect.transformed(t).rectified()


def iterate_vector(position, t):
    return t.dot(position)


def test_projection():
    space_converter = OrthoProjection(Rect([-10.0, -15.0],
                                           [40.0, 45.0]))

    def start_position_f(space_converter, space):
        return four_components(np.random.uniform(size=2))
    process(space_converter, start_position_f, iterate_vector, 2)


def test_view():
    space_converter = OrthoView(zoom=[2.4, 31.2],
                                scroll=[8, 3])
    space_converter.device_pixel_ratio = 3.0

    def start_position_f(space_converter, space):
        return four_components(np.random.uniform(size=2))
    process(space_converter, start_position_f, iterate_vector, 39)



__all__ = ['Attribute']


class Attribute:

    def __init__(self,
                 name,
                 lookup_sequence,
                 *,
                 is_vector=False,
                 is_packed_array=False,
                 array_size=None):
        """
        * name is the name of the input attribute in the GLSL vertex
          shader code.  It is mainly used for error-reporting.
        * lookup_sequence is an iterable of names or indices, which constitute
          a path in the numpy array's dtype that locates the data in the numpy
          array.  Names are used to index into records, and indices are used
          to lookup into arrays.
        * is_vector is a Boolean that says whether this attribute is a vector,
          e.g., vec4.  Vector attributes have between 2 and 4 components.  The
          final dimension of the numpy array's shape must match the vector
          length.
        * is_packed_array is a Boolean that indicates that the data is packed
          as vectors of length 4, which means that the array_size times 4 is
          greater than or equal to the final dimension of the numpy array's
          shape.
        * array_size: an integer representing the size of the array
          attribute, e.g., float data[7].  For now, these must be bound to
          contiguous entries in the buffer numpy array.  Otherwise
          array_size is None to indicate a scalar.
        """
        self.name = name
        self.lookup_sequence = lookup_sequence
        self.is_vector = is_vector
        self.array_size = array_size
        self.is_packed_array = is_packed_array
        if is_vector and is_packed_array:
            raise ValueError

    def __repr__(self):
        return (f"{type(self).__qualname__}("
                f"{self.name}, {self.lookup_sequence}, "
                f"is_vector={self.is_vector}, "
                f"is_packed_array={self.is_packed_array}, "
                f"array_size={self.array_size})")



from contextlib import contextmanager

import numpy as np

from ..gl_importer import gl as gl

__all__ = ['Atlas']


class Atlas:

    class OverfullError(BaseException):
        pass

    class Line:

        def __init__(self, y):
            """
            (x, y) is the top-left corner.
            """
            self.y = y
            self.x = 0
            self.height = 0

    def __init__(self, font, size, texture_unit, guard=1):
        self.creation_id = 0
        self.clear()
        self.resize(size)
        self.texture_unit = texture_unit
        self.font = font
        self.guard = guard

        with self.font.shader_program.bind_context():
            # Create sampler.
            self.sampler_object, = gl.glGenSamplers(1)

            # Create texture.
            self.texture, = gl.glGenTextures(1)

            # Bind sampler to texture unit, and set parameters.
            gl.glActiveTexture(gl.GL_TEXTURE0 + self.texture_unit)
            gl.glBindSampler(self.texture_unit, self.sampler_object)
            gl.glSamplerParameteri(self.sampler_object,
                                   gl.GL_TEXTURE_WRAP_S,
                                   gl.GL_CLAMP_TO_EDGE)
            gl.glSamplerParameteri(self.sampler_object,
                                   gl.GL_TEXTURE_WRAP_T,
                                   gl.GL_CLAMP_TO_EDGE)
            gl.glSamplerParameteri(self.sampler_object,
                                   gl.GL_TEXTURE_MAG_FILTER,
                                   gl.GL_LINEAR)
            gl.glSamplerParameteri(self.sampler_object,
                                   gl.GL_TEXTURE_MIN_FILTER,
                                   gl.GL_LINEAR)
            gl.glBindSampler(self.texture_unit, 0)

            # Set uniform with texture unit.
            self.font.shader_program.font_atlas(np.int32(self.texture_unit))

    def resize(self, size):
        self.creation_id += 1
        self.needs_update = True
        self.texture_array = np.empty((size, size), dtype=np.ubyte)

    def clear(self):
        self.needs_update = True
        self.lines = []
        self.y = 0  # rows used

    @property
    def size(self):
        return self.texture_array.shape[0]

    def new_line(self):
        if self.lines:
            self.y += self.lines[-1].height + self.guard
        line = Atlas.Line(self.y)
        self.lines.append(line)
        return line

    def add_char(self, bitmap):
        while True:
            try:
                if bitmap.width > self.size:
                    raise Atlas.OverfullError
                for line in self.lines:
                    if (bitmap.width + line.x <= self.size
                            and (bitmap.rows <= line.height
                                 or line is self.lines[-1])):
                        break
                else:
                    line = self.new_line()
                    if self.y + bitmap.rows > self.size:
                        raise Atlas.OverfullError

                # Copy bitmap into texture_array.
                self.texture_array[line.y: line.y + bitmap.rows,
                                   line.x: line.x + bitmap.width].flat = \
                    bitmap.buffer
                self.needs_update = True

                line.height = max(bitmap.rows, line.height)
                bottom_left = [line.x, line.y + bitmap.rows]
                top_right = [line.x + bitmap.width, line.y]
                line.x += bitmap.width + self.guard
                return np.array(bottom_left + top_right) / self.size
            except Atlas.OverfullError:
                self.resize(self.size * 2)
                self.font.repopulate()

    def update_texture(self):
        # Bind texture to texture unit, set paramters and upload texture.
        # ActiveTexture must precede TexParameter, BindTexture,
        # and TexImage.
        gl.glActiveTexture(gl.GL_TEXTURE0 + self.texture_unit)
        gl.glBindTexture(gl.GL_TEXTURE_2D, self.texture)
        gl.glTexImage2D(gl.GL_TEXTURE_2D,
                        0,
                        gl.GL_R8,
                        self.texture_array.shape[1],
                        self.texture_array.shape[0],
                        0,
                        gl.GL_RED,
                        gl.GL_UNSIGNED_BYTE,
                        self.texture_array)
        self.needs_update = False

    @contextmanager
    def draw_context(self):
        gl.glActiveTexture(gl.GL_TEXTURE0 + self.texture_unit)
        gl.glBindTexture(gl.GL_TEXTURE_2D, self.texture)
        gl.glBindSampler(self.texture_unit, self.sampler_object)
        yield
        gl.glActiveTexture(gl.GL_TEXTURE0 + self.texture_unit)
        gl.glBindTexture(gl.GL_TEXTURE_2D, 0)
        gl.glBindSampler(self.texture_unit, 0)



import numpy as np
from pkg_resources import resource_filename

from ..gl_importer import gl
from ..shader_program import Attribute, BufferDescription, ShaderProgram

__all__ = ['BasicShaderProgram']


class BasicShaderProgram(ShaderProgram):

    """
    The BasicShaderProgram draws pairs of (x, y) points, as for example
    triangle strips or lines.
    """

    def __init__(self, uniforms={}):
        super().__init__(
            vertex=[resource_filename('glx', 'glsl_shaders/basic.vert')],
            fragment=[resource_filename('glx', 'glsl_shaders/basic.frag')])

        with self.bind_context(uniforms):
            # Create one buffer.
            self.buffer, = gl.glGenBuffers(1)

            # Create one vertex array and describe it.
            # The array is fed using a numpy array of shape (n, 2), which
            # describes n (x, y) pairs.
            self.vertex_array, = self.create_vertex_arrays(
                [BufferDescription(
                    self.buffer,
                    np.dtype(('f', (2,))),
                    [Attribute('vertex', [], is_vector=True)])])



#!/usr/bin/env python
from setuptools import find_packages, setup

setup(
    name='glx',
    version='0.5.0',
    description=('Classes that facilitate working with OpenGL 4.4 shaders '
                 'and efficiently rendering text.'),
    author='Neil Girdhar',
    author_email='mistersheik@gmail.com',
    url='https://github.com/NeilGirdhar/glx',
    download_url='https://github.com/neilgirdhar/glx/archive/0.5.0.tar.gz',
    packages=find_packages(),
    package_data={
        'glx': ['**/*.vert', '**/*.frag', '**/*.geom'],
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],
    keywords=['OpenGL'],
    install_requires=['numpy>=1.13',
                      'mako>=1.0.7',
                      'rectangle>=0.2',
                      'freetype-py>=1.1',
                      'PyOpenGL>=3.0.0'],
    python_requires='>=3.4',
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
)



from contextlib import contextmanager

import numpy as np
from rectangle import Rect

from .ortho_view import OrthoView

__all__ = ['BoundedOrthoView']


class BoundedOrthoView(OrthoView):
    """
    Like an OrthoView, BoundedOrthoView converts between four spaces:
    * scene
    * screen (the scene space zoomed)
    * widget (the screen space translated)
    * scaled_widget (device_pixel_ratio times smaller than widget)
    It also keeps track of
    * the size of the widget in screen space,
    * the represented area in scene space,
    * a range of valid zoom levels, and
    * a border.
    """

    def __init__(self,
                 zoom,
                 scroll,
                 scene_rect,
                 border,
                 zoom_range,
                 widget_size):
        # pylint: disable=too-many-arguments
        """
        * zoom
          * shape: (2,)
          * Conversion from scene space to screen space.
        * scroll
          * shape: (2,)
          * Conversion from screen space to widget space.
          * Bottom-right is positive.
        * scene_rect
          * Rect of shape: (2,)
          * The rectangle of the scene that is represented.
          * in scene space
        * border
          * float
          * The border in pixels (screen space).
        * zoom_range
          * shape: (2,)
          * The minimum and maximum zoom for both axes.
          * Limits the zoom whenever it's set.
        * widget_size
          * shape: (2,)
          * The size of the widget in screen space.
          * Limits the scroll whenever it's set.
        """
        super().__init__(zoom=zoom, scroll=scroll)
        self.scene_rect_ = scene_rect
        self.border = border
        self.zoom_range = zoom_range
        self.widget_size_ = widget_size

    # Settable vector properties ----------------------------------------------
    @OrthoView.zoom.setter
    def zoom(self, new_zoom):
        if not isinstance(new_zoom, np.ndarray):
            raise TypeError
        # Clip the new_zoom to zoom_range
        OrthoView.zoom.fset(self, np.clip(new_zoom, *self.zoom_range))

    @OrthoView.scroll.setter
    def scroll(self, new_scroll):
        OrthoView.scroll.fset(self, self.scroll_range_rect.clamped(new_scroll))

    @property
    def scene_rect(self):
        return self.scene_rect_

    @scene_rect.setter
    def scene_rect(self, new_scene_rect):
        self.scene_rect_ = new_scene_rect

    @property
    def widget_size(self):
        return self.widget_size_

    @widget_size.setter
    def widget_size(self, new_widget_size):
        self.widget_size_ = new_widget_size
        self.scroll = self.scroll  # Clamp the scroll.

    # Rectangle properties ----------------------------------------------------
    @property
    def screen_rect(self):
        return (self.scene_rect * self.flipped_zoom).rectified().bordered(
            self.border)

    @property
    def scroll_range_rect(self):
        return Rect(self.screen_rect.mins,
                    self.screen_rect.maxes - self.widget_size)

    # Value updaters ----------------------------------------------------------
    def hold_and_set_scene_rect(self,
                                widget_hold_point,
                                scene_rect):
        with self.hold_position(widget_hold_point):
            self.scene_rect = scene_rect

    @contextmanager
    def hold_position(self, widget_hold_point):
        """
        Scrolls the scene to make a point (in widget space) map to the same
        scene space.  This keeps the spot under the pointer from moving when
        the zoom command is being issued with the mouse.
        """
        def scene_hold_point(widget_hold_point):
            if widget_hold_point is None:
                return np.zeros(4)
            return self.widget_to_scene.dot(widget_hold_point)
        shp = scene_hold_point(widget_hold_point)
        try:
            yield
        finally:
            delta_shp = (shp - scene_hold_point(widget_hold_point))[:2]
            self.scroll += (delta_shp * self.flipped_zoom).astype(np.int32)

    def __repr__(self):
        return (f"{type(self).__name__}("
                f"zoom={self.zoom}, "
                f"scroll={self.scroll}, "
                f"scene_rect={self.scene_rect}, "
                f"border={self.border}, "
                f"zoom_range={self.zoom_range}, "
                f"widget_size={self.widget_size}"
                ")")



import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal, assert_array_equal

from .transformations import clip_matrix


@pytest.fixture
def setup_numpy(scope='module'):
    np.set_printoptions(suppress=True, precision=5)
    yield


def test_clip_matrix(random_state):
    frustum = random_state.random_sample(6)
    frustum[1] += frustum[0]
    frustum[3] += frustum[2]
    frustum[5] += frustum[4]

    cm = clip_matrix(perspective=False, *frustum)
    assert_array_equal(
        np.dot(cm, [frustum[0], frustum[2], frustum[4], 1]),
        [-1., -1., -1., 1.])
    assert_array_equal(
        np.dot(cm, [frustum[1], frustum[3], frustum[5], 1]),
        [1., 1., 1., 1.])

    cm = clip_matrix(perspective=True, *frustum)
    v = np.dot(cm, [frustum[0], frustum[2], frustum[4], 1])
    assert_array_almost_equal(v / v[3],
                              [-1., -1., -1., 1.])
    v = np.dot(cm, [frustum[1], frustum[3], frustum[4], 1])
    assert_array_almost_equal(v / v[3],
                              [1., 1., -1., 1.])



from contextlib import contextmanager

from ..gl_importer import OpenGL, gl
from ..wrappers import glGetActiveAttrib
from .shader import Shader
from .uniform_description import UniformDescription
from .vertex_array import VertexArray

__all__ = ['ShaderProgram']


class ShaderProgram:

    """
    A ShaderProgram manages an entire pipeline: vertex shader, possible
    geometry shader, and fragment shader.  Calling bind_context binds the
    entire pipeline, and possible some uniform variables.   Its
    create_vertex_arrays method creates a list of VertexArray objects, which
    allows binding of varying variables.
    """

    def __init__(self, vertex=[], geometry=[], fragment=[],
                 context_kwargs=None):
        """
        * vertex, geometry and fragment are lists of filenames of included
          shaders.
        * context_kwargs are passed to the mako runtime context.
        """
        # pylint: disable=assignment-from-no-return
        self.program_index = gl.glCreateProgram()
        assert self.program_index > 0
        self.shaders = {(filename, type_): Shader(filename,
                                                  type_,
                                                  context_kwargs)
                        for filenames, type_ in [
                            (vertex, gl.GL_VERTEX_SHADER),
                            (geometry, gl.GL_GEOMETRY_SHADER),
                            (fragment, gl.GL_FRAGMENT_SHADER)]
                        for filename in filenames}
        for shader in self.shaders.values():
            gl.glAttachShader(self.program_index,
                              shader.shader_index)
        gl.glLinkProgram(self.program_index)
        value = gl.glGetProgramiv(self.program_index, gl.GL_LINK_STATUS)
        if value != 1:
            log = gl.glGetProgramInfoLog(self.program_index).decode('latin')
            raise Exception("""
                            Couldn't link program.
                            Shader program info log:
                            """ + log)

        self.create_uniform_binders()

    @contextmanager
    def bind_context(self, uniforms=None):
        if uniforms is None:
            uniforms = {}

        try:
            gl.glUseProgram(self.program_index)
        except OpenGL.error.GLError:
            log = gl.glGetProgramInfoLog(self.program_index).decode('latin')
            raise Exception("""
                            Couldn't use program.
                            Shader program info log:
                            """ + log)
        try:
            for name, value in uniforms.items():
                getattr(self, name)(value)
            yield self
        finally:
            gl.glUseProgram(0)

    def delete_program(self):
        gl.glDeleteProgam(self.program_index)
        self.program_index = 0

    def create_uniform_binders(self):
        for shader in self.shaders.values():
            if shader.type_ != gl.GL_VERTEX_SHADER:
                continue
            for u in UniformDescription.parse_shader_text(shader.shader_text):
                if hasattr(self, u.name):
                    raise ValueError(
                        f"Multiple uniforms with the same name: {u.name}")
                setattr(self, u.name, u.create_method(self.program_index))

    def attribute_name_to_location(self, attribute_name):
        attribute_location = gl.glGetAttribLocation(
            self.program_index, attribute_name)
        if attribute_location == -1:
            count = gl.glGetProgramiv(self.program_index,
                                      gl.GL_ACTIVE_ATTRIBUTES)[0]
            print("Active Attributes:")
            for i in range(count):
                name, _, _ = glGetActiveAttrib(self.program_index, i)
                print(" ", name)
            raise ValueError(
                f'Attribute "{attribute_name}" not found in program.')
        return attribute_location

    def create_vertex_arrays(self, buffer_descriptions):
        """
        * buffer_descriptions is a reiterable of BufferDescription.
        """
        return [VertexArray(self, buffer_description)
                for buffer_description in buffer_descriptions]
